package tech.sbdevelopment.vehiclesplusconverter.utils;

import com.google.common.io.Files;
import net.md_5.bungee.api.ChatColor;
import nl.sbdeveloper.vehiclesplus.storage.file.HJSONFile;
import org.bukkit.inventory.ItemStack;
import org.bukkit.plugin.java.JavaPlugin;
import tech.sbdevelopment.vehiclesplusconverter.VehiclesPlusConverter;
import tech.sbdevelopment.vehiclesplusconverter.api.ConversionException;
import tech.sbdevelopment.vehiclesplusconverter.api.InvalidConversionException;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.logging.Level;

public class MainUtil {
    private MainUtil() {
    }

    public static String __(String in) {
        return ChatColor.translateAlternateColorCodes('&', in);
    }

    public static String getClassByFullName(String name) {
        String[] split = name.split("\\.");
        if (split.length == 0) return name;
        return split[split.length - 1]; //Last position
    }

    public static String idToReadable(String id) {
        if (id.contains("_")) {
            // Split on _ and capitalize each part
            String[] split = id.split("_");
            StringBuilder builder = new StringBuilder();
            for (String s : split) {
                builder.append(s.substring(0, 1).toUpperCase()).append(s.substring(1));
                builder.append(" ");
            }
            return builder.toString().trim();
        } else {
            // Capitalize the first letter
            return id.substring(0, 1).toUpperCase() + id.substring(1);
        }
    }

    public static void disablePlugin(JavaPlugin plugin) throws InvocationTargetException, IllegalAccessException, NoSuchMethodException, IOException {
        // Call protected .getFile() method, then rename that file to .jar.old and save
        Method getFile = JavaPlugin.class.getDeclaredMethod("getFile");
        getFile.setAccessible(true);
        File file = (File) getFile.invoke(plugin);
        File renamedFile = new File(file.getParent(), file.getName() + ".disabled");
        Files.move(file, renamedFile);
    }

    public static String getTypeIdByClass(String baseVehicle, String type) throws ConversionException {
        switch (type) {
            case "BikeType":
                return "bikes";
            case "BoatType":
                return "boats";
            case "CarType":
                return "cars";
            case "HelicopterType":
                return "helicopters";
            case "HovercraftType":
                return "hovercrafts";
            case "PlaneType":
                return "planes";
            default:
                throw new InvalidConversionException("vehicleType", baseVehicle);
        }
    }

    public static void saveToVehiclesPlus(Object data, String subFolder, String fileName) {
        File parentFolders = new File(nl.sbdeveloper.vehiclesplus.VehiclesPlus.getInstance().getDataFolder(), subFolder);
        if (!parentFolders.exists() && !parentFolders.mkdirs()) {
            VehiclesPlusConverter.getInstance().getLogger().log(Level.SEVERE, "Couldn't create the folder " + subFolder);
            return;
        }

        HJSONFile jsonFile = new HJSONFile(nl.sbdeveloper.vehiclesplus.VehiclesPlus.getInstance(), subFolder + "/" + fileName);
        try {
            jsonFile.write(data);
        } catch (IOException e) {
            VehiclesPlusConverter.getInstance().getLogger().log(Level.SEVERE, "Couldn't save to the file " + fileName, e);
        }
    }

    /**
     * Compare two ItemStacks for equality based on material type and metadata
     *
     * @param item1 First ItemStack
     * @param item2 Second ItemStack
     * @return true if the items are considered equal
     */
    public static boolean compareItems(ItemStack item1, ItemStack item2) {
        if (item1 == null || item2 == null) return false;
        if (!item1.getType().equals(item2.getType())) return false;

        org.bukkit.inventory.meta.ItemMeta meta1 = item1.getItemMeta();
        org.bukkit.inventory.meta.ItemMeta meta2 = item2.getItemMeta();

        if (meta1 == null || meta2 == null) return meta1 == meta2;

        // Check CustomModelData if available (1.14+)
        try {
            return meta1.hasCustomModelData() && meta2.hasCustomModelData()
                    && meta1.getCustomModelData() == meta2.getCustomModelData();
        } catch (NoSuchMethodError ignored) {
            // Pre 1.14, check Durability/Damage value instead
            return item1.getDurability() == item2.getDurability();
        }
    }

    private static final String[] RIM_PREFIXES = {
            "sport", "racing", "luxury", "classic", "vintage", "modern", "elite", "premium", "pro", "ultra"
    };

    private static final String[] RIM_TYPES = {
            "star", "spoke", "mesh", "alloy", "chrome", "forged", "cast", "split", "blade", "wave"
    };

    private static final String[] RIM_SUFFIXES = {
            "design", "series", "line", "edition", "collection", "style", "plus", "max", "elite", "pro"
    };

    /**
     * Generate a realistic rim design name
     *
     * @return A random rim design name
     */
    public static String generateRimName() {
        String prefix = RIM_PREFIXES[(int) (Math.random() * RIM_PREFIXES.length)];
        String type = RIM_TYPES[(int) (Math.random() * RIM_TYPES.length)];
        String suffix = RIM_SUFFIXES[(int) (Math.random() * RIM_SUFFIXES.length)];

        // Add a random number between 100 and 999
        int number = 100 + (int) (Math.random() * 900);

        // Convert to PascalCase
        return prefix.substring(0, 1).toUpperCase() + prefix.substring(1) +
                type.substring(0, 1).toUpperCase() + type.substring(1) +
                suffix.substring(0, 1).toUpperCase() + suffix.substring(1) +
                number;
    }
}
